#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-12
# @Filename: specs.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, cast, get_args

from typing_extensions import TypedDict

from sdsstools.utils import GatheringTaskGroup

from lvmopstools.clu import CluClient


if TYPE_CHECKING:
    from clu.command import Command


__all__ = [
    "spectrograph_temperature_label",
    "spectrograph_temperatures",
    "spectrograph_pressures",
    "spectrograph_mechanics",
    "spectrograph_status",
    "exposure_etr",
]


Spectrographs = Literal["sp1", "sp2", "sp3"]
Cameras = Literal["r", "b", "z"]
Sensors = Literal["ccd", "ln2"]
SpecStatus = Literal["idle", "exposing", "reading", "error", "unknown"]
SpecToStatus = dict[Spectrographs, SpecStatus]


def spectrograph_temperature_label(camera: str, sensor: str = "ccd") -> str:
    """Returns the archon label associated with a temperature sensor."""

    if sensor == "ccd":
        if camera == "r":
            return "mod2/tempa"
        elif camera == "b":
            return "mod12/tempc"
        elif camera == "z":
            return "mod12/tempa"

    else:
        if camera == "r":
            return "mod2/tempb"
        elif camera == "b":
            return "mod2/tempc"
        elif camera == "z":
            return "mod12/tempb"

    raise ValueError(f"Invalid camera {camera!r} or sensor {sensor!r}.")


async def spectrograph_temperatures(
    spec: Spectrographs | None = None,
    ignore_errors: bool = True,
) -> dict[str, float | None]:
    """Returns a dictionary of spectrograph temperatures.

    Parameters
    ----------
    spec
        The spectrograph to retrieve the temperatures for. If `None`, retrieves
        the temperatures for all spectrographs.
    ignore_errors
        If `True`, ignores errors when retrieving the temperatures and replaces the
        missing values with `None`. If `False`, raises an error if any of the
        temperatures cannot be retrieved.

    Returns
    -------
    dict
        A dictionary with the temperatures for each camera and sensor, e.g.,
        ``{'r1_ln2': -184.1, 'r1_ccd': -120.3, ...}``.

    """

    if spec is None:
        async with GatheringTaskGroup() as group:
            for spec in get_args(Spectrographs):
                group.create_task(
                    spectrograph_temperatures(
                        spec,
                        ignore_errors=ignore_errors,
                    )
                )

        return {
            key: value
            for task_result in group.results()
            for key, value in task_result.items()
        }

    if spec not in get_args(Spectrographs):
        raise ValueError(f"Invalid spectrograph {spec!r}.")

    async with CluClient() as client:
        scp_command = await client.send_command(
            f"lvmscp.{spec}",
            "status",
            internal=True,
        )

    try:
        if scp_command.status.did_fail:
            raise ValueError(f"Failed retrieving status from SCP for spec {spec!r}.")

        status = scp_command.replies.get("status")
    except Exception:
        if not ignore_errors:
            raise

        status = {}

    response: dict[str, float | None] = {}

    cameras: list[Cameras] = ["r", "b", "z"]
    sensors: list[Sensors] = ["ccd", "ln2"]

    for camera in cameras:
        for sensor in sensors:
            label = spectrograph_temperature_label(camera, sensor)
            if label not in status:
                if not ignore_errors:
                    raise ValueError(f"Cannot find status label {label!r}.")
                else:
                    value = None
            else:
                value = status[label]

            response[f"{camera}{spec[-1]}_{sensor}"] = value

    return response


async def spectrograph_pressures(
    spec: Spectrographs | None = None,
    ignore_errors: bool = True,
) -> dict[str, float | None]:
    """Returns a dictionary of spectrograph pressures.

    Parameters
    ----------
    spec
        The spectrograph to retrieve the pressures for. If `None`, retrieves
        the pressures for all spectrographs.
    ignore_errors
        If `True`, ignores errors when retrieving the pressures and replaces the
        missing values with `None`. If `False`, raises an error if any of the
        pressures cannot be retrieved.

    """

    if spec is None:
        async with GatheringTaskGroup() as group:
            for spec in get_args(Spectrographs):
                group.create_task(
                    spectrograph_pressures(
                        spec,
                        ignore_errors=ignore_errors,
                    )
                )

        return {
            key: value
            for task_result in group.results()
            for key, value in task_result.items()
        }

    if spec not in get_args(Spectrographs):
        raise ValueError(f"Invalid spectrograph {spec!r}.")

    async with CluClient() as client:
        ieb_command = await client.send_command(
            f"lvmieb.{spec}",
            "transducer status",
            internal=True,
        )

    try:
        if ieb_command.status.did_fail:
            raise ValueError("Failed retrieving status from IEB.")
        pressures = ieb_command.replies.get("transducer")
    except Exception:
        if not ignore_errors:
            raise

        pressures = {}

    response: dict[str, float | None] = {}

    spec_id = spec[-1]
    keys = [f"{camera}{spec_id}_pressure" for camera in get_args(Cameras)]
    for key in keys:
        camera = key.split("_")[0]
        if key in pressures:
            response[camera] = pressures[key]
        else:
            if not ignore_errors:
                raise ValueError(f"Cannot find pressure in key {key!r}.")
            else:
                response[camera] = None

    return response


async def spectrograph_mechanics(
    spec: Spectrographs | None = None,
    ignore_errors: bool = True,
) -> dict[str, str | None]:
    """Returns a dictionary of spectrograph shutter and hartmann door status.

    Parameters
    ----------
    spec
        The spectrograph to retrieve the mechanics status for. If `None`, retrieves
        the status for all spectrographs.
    ignore_errors
        If `True`, ignores errors when retrieving the status and replaces the
        missing values with `None`. If `False`, raises an error if any of the
        status cannot be retrieved

    """

    def get_reply(cmd: Command, key: str):
        try:
            return "open" if cmd.replies.get(key)["open"] else "closed"
        except Exception:
            if not ignore_errors:
                raise ValueError(f"Cannot find key {key!r} in IEB command replies.")
            else:
                return None

    if spec is None:
        async with GatheringTaskGroup() as group:
            for spec in get_args(Spectrographs):
                group.create_task(
                    spectrograph_mechanics(
                        spec,
                        ignore_errors=ignore_errors,
                    )
                )

        return {
            key: value
            for task_result in group.results()
            for key, value in task_result.items()
        }

    if spec not in get_args(Spectrographs):
        raise ValueError(f"Invalid spectrograph {spec!r}.")

    response: dict[str, str | None] = {}

    async with CluClient() as client:
        for device in ["shutter", "hartmann"]:
            ieb_cmd = await client.send_command(
                f"lvmieb.{spec}",
                f"{device} status",
                internal=True,
            )

            if ieb_cmd.status.did_fail:
                if not ignore_errors:
                    raise ValueError(f"Failed retrieving {device} status from IEB.")

            if device == "shutter":
                key = f"{spec}_shutter"
                response[key] = get_reply(ieb_cmd, key)
            else:
                for door in ["left", "right"]:
                    key = f"{spec}_hartmann_{door}"
                    response[key] = get_reply(ieb_cmd, key)

    return response


async def exposure_etr() -> tuple[float | None, float | None]:
    """Returns the ETR for the exposure, including readout."""

    spec_names = get_args(Spectrographs)

    async with CluClient() as client:
        async with GatheringTaskGroup() as group:
            for spec in spec_names:
                group.create_task(
                    client.send_command(
                        f"lvmscp.{spec}",
                        "get-etr",
                        internal=True,
                    )
                )

    etrs: list[float] = []
    total_times: list[float] = []
    for task in group.results():
        if task.status.did_fail:
            continue

        etr = task.replies.get("etr")
        if all(etr):
            etrs.append(etr[0])
            total_times.append(etr[1])

    if len(etrs) == 0 or len(total_times) == 0:
        return None, None

    return max(etrs), max(total_times)


class SpectrographStatusResponse(TypedDict):
    """Spectrograph status response."""

    status: SpecToStatus
    last_exposure_no: int
    etr: tuple[float, float] | tuple[None, None]


async def spectrograph_status() -> SpectrographStatusResponse:
    """Returns the status of the spectrographs."""

    spec_names = get_args(Spectrographs)

    async with CluClient() as client:
        async with GatheringTaskGroup() as group:
            for spec in spec_names:
                group.create_task(
                    client.send_command(
                        f"lvmscp.{spec}",
                        "status -s",
                        internal=True,
                    )
                )
            group.create_task(exposure_etr())

    status_dict: SpecToStatus = {}
    last_exposure_no: int = -1
    etr: tuple[float, float] | tuple[None, None] = (None, None)

    for itask, task in enumerate(group.results()):
        if itask == len(spec_names):
            etr = cast(tuple[float, float] | tuple[None, None], task)
            continue

        if task.status.did_fail:
            continue

        status = task.replies.get("status")
        controller: Spectrographs = status["controller"]
        status_names: str = status["status_names"]

        if "ERROR" in status_names:
            status_dict[controller] = "error"
        elif "IDLE" in status_names:
            status_dict[controller] = "idle"
        elif "EXPOSING" in status_names:
            status_dict[controller] = "exposing"
        elif "READING" in status_names:
            status_dict[controller] = "reading"
        else:
            status_dict[controller] = "unknown"

        last_exposure_no_key = status.get("last_exposure_no", -1)
        if last_exposure_no_key > last_exposure_no:
            last_exposure_no = cast(int, last_exposure_no_key)

    for spec in spec_names:
        if spec not in status_dict:
            status_dict[spec] = "unknown"

    response: SpectrographStatusResponse = {
        "status": status_dict,
        "last_exposure_no": last_exposure_no,
        "etr": etr,
    }

    return response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-13
# @Filename: nps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Sequence

from typing_extensions import TypedDict

from sdsstools import GatheringTaskGroup

from lvmopstools import config
from lvmopstools.clu import send_clu_command
from lvmopstools.retrier import Retrier


__all__ = ["read_nps", "read_outlet"]


class NPSStatus(TypedDict):
    """A class to represent the status of an NPS."""

    actor: str
    name: str
    id: int
    state: bool


@Retrier(max_attempts=3, delay=1)
async def read_nps(actors: Sequence[str] | str | None = None) -> dict[str, NPSStatus]:
    """Returns the status of all NPS."""

    default_actors: list[str] = config["devices.nps.default_actors"]
    valid_actors: list[str] = config["devices.nps.valid_actors"]

    if isinstance(actors, str):
        actors = [actors]

    if actors is None:
        actors = default_actors

    async with GatheringTaskGroup() as group:
        for actor in actors:
            if not actor.startswith("lvmnps."):
                actor = "lvmnps." + actor

            if actor not in valid_actors:
                raise ValueError(f"Invalid NPS actor: {actor}.")

            group.create_task(
                send_clu_command(
                    f"{actor} status",
                    raw=True,
                    internal=True,
                )
            )

    nps_data: dict[str, NPSStatus] = {}

    for cmd in group.results():
        actor = cmd.consumer_id
        outlets = cmd.replies.get("outlets")
        for outlet in outlets:
            key = f"{actor.split('.')[1]}.{outlet['normalised_name']}"
            nps_data[key] = NPSStatus(
                actor=actor,
                name=outlet["normalised_name"],
                id=outlet["id"],
                state=outlet["state"],
            )

    return nps_data


@Retrier(max_attempts=3, delay=1)
async def read_outlet(actor: str, outlet: str | int) -> NPSStatus:
    """Returns the status of a single NPS outlet."""

    data = await send_clu_command(f"{actor} status {outlet}", internal=True)

    return {
        "actor": actor,
        "name": data[0]["outlet_info"]["normalised_name"],
        "id": data[0]["outlet_info"]["id"],
        "state": data[0]["outlet_info"]["state"],
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2025-08-10
# @Filename: ags.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from lvmopstools import config
from lvmopstools.clu import send_clu_command
from lvmopstools.devices.switch import power_cycle_interface


async def power_cycle_ag_camera_nps(
    camera: str,
    delay: float = 3,
    verbose: bool = False,
):
    """Power cycles the NPS for the given auto-guide camera.

    Parameters
    ----------
    camera
        The name of the camera to power cycle.
    delay
        The delay in seconds before power cycling the camera.
    verbose
        If True, prints the commands that are being run.

    """

    ags_config = config["devices.agcams.power"][camera]
    actor = ags_config["actor"]
    outlet = ags_config["outlet"]

    if verbose:
        print(f"Power cycling camera {camera!r} via NPS...")

    cmd = await send_clu_command(f"{actor} cycle --delay {delay} {outlet}", raw=True)
    if cmd.status.did_fail:
        raise RuntimeError(f"Failed to power cycle camera {camera!r} via NPS.")

    if verbose:
        print(f"Power cycling of {camera!r} complete.")


async def power_cycle_ag_camera(camera: str, verbose: bool = False):
    """Power cycles an AG camera either by resetting PoE or toggling the NPS.

    Parameters
    ----------
    camera
        The name of the camera to power cycle.
    verbose
        If True, prints the commands that are being run.

    """

    if not isinstance(camera, str):
        camera = str(camera)

    ags_config = config["devices.agcams.power"]

    for cam_config_name in ags_config:
        aliases = ags_config[cam_config_name]["aliases"] or []
        if camera.lower() == cam_config_name.lower() or camera in aliases:
            camera = cam_config_name
            break
    else:
        raise ValueError(f"Invalid camera name {camera}")

    ag_camera_config = ags_config[camera]

    if ag_camera_config["mode"] == "poe":
        interface = ag_camera_config["interface"]
        power_cycle_interface(interface, verbose=verbose)
    elif ag_camera_config["mode"] == "nps":
        await power_cycle_ag_camera_nps(camera, verbose=verbose)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-02
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib
import warnings

from sdsstools import Configuration, read_yaml_file
from sdsstools.metadata import get_package_version


__version__ = get_package_version(path=__file__, package_name="lvmopstools")


DEFAULT_CONFIG_FILE = pathlib.Path(__file__).parent / "data" / "config.yaml"
CONFIG_FILE = pathlib.Path(
    os.environ.get("LVMOPSTOOLS_CONFIG_FILE", DEFAULT_CONFIG_FILE)
)

if not CONFIG_FILE.exists():
    warnings.warn(
        f"Config file not found at {CONFIG_FILE!s}. "
        "Reverting to internal configuration."
    )
    CONFIG_FILE = DEFAULT_CONFIG_FILE


config = read_yaml_file(CONFIG_FILE, return_class=Configuration)


def set_config(config_file: str | pathlib.Path | None = None) -> None:
    """Sets the configuration file."""

    global CONFIG_FILE

    config_path = pathlib.Path(config_file or DEFAULT_CONFIG_FILE)
    if not config_path.exists():
        raise FileNotFoundError(f"Config file {config_path!s} not found.")

    config.load(config_path, use_base=False)

    CONFIG_FILE = config_path


from .retrier import Retrier
from .socket import AsyncSocketHandler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-21
# @Filename: notifications.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import enum
import pathlib
import smtplib
import sys
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from typing import Any, Sequence, cast

from jinja2 import Environment, FileSystemLoader

from lvmopstools import config
from lvmopstools.slack import post_message as post_to_slack


__all__ = [
    "send_notification",
    "send_email",
    "send_critical_error_email",
    "NotificationLevel",
]


class NotificationLevel(enum.Enum):
    """Allowed notification levels."""

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


async def send_notification(
    message: str,
    level: NotificationLevel | str = NotificationLevel.INFO,
    slack: bool = True,
    slack_channels: str | Sequence[str] | None = None,
    email_on_critical: bool = True,
    slack_extra_params: dict[str, Any] = {},
    email_params: dict[str, Any] = {},
):
    """Creates a new notification.

    Parameters
    ----------
    message
        The message of the notification. Can be formatted in Markdown.
    level
        The level of the notification.
    slack
        Whether to send the notification to Slack.
    slack_channels
        The Slack channel where to send the notification. If not provided, the default
        channel is used. Can be set to false to disable sending the Slack notification.
    email_on_critical
        Whether to send an email if the notification level is ``CRITICAL``.
    slack_extra_params
        A dictionary of extra parameters to pass to ``post_message``.
    email_params
        A dictionary of extra parameters to pass to :obj:`.send_critical_error_email`.

    Returns
    -------
    message
        The message that was sent.

    """

    if isinstance(level, str):
        level = NotificationLevel(level.upper())
    else:
        level = NotificationLevel(level)

    send_email = email_on_critical and level == NotificationLevel.CRITICAL

    if send_email:
        try:
            send_critical_error_email(message, **email_params)
        except Exception as ee:
            print(f"Error sending critical error email: {ee}", file=sys.stderr)

    if slack:
        slack_channels = slack_channels or config["slack.default_channels"]

        channels: set[str] = set()

        if isinstance(slack_channels, str):
            channels.add(slack_channels)
        elif isinstance(slack_channels, Sequence):
            channels.update(slack_channels)

        # We send the message to the default channel plus any other channel that
        # matches the level of the notification.
        level_channels = cast(dict[str, str], config["slack.level_channels"])
        if level.value in level_channels:
            if isinstance(level_channels[level.value], str):
                channels.add(level_channels[level.value])
            else:
                channels.update(level_channels[level.value])

        # Send Slack message(s)
        for channel in channels:
            mentions = (
                ["@channel"]
                if level == NotificationLevel.CRITICAL
                or level == NotificationLevel.ERROR
                else []
            )
            try:
                await post_to_slack(
                    message,
                    channel=channel,
                    mentions=mentions,
                    **slack_extra_params,
                )
            except Exception as se:
                print(f"Error sending Slack message: {se}", file=sys.stderr)

    return message


def send_email(
    message: str,
    subject: str,
    recipients: Sequence[str],
    from_address: str,
    *,
    html_message: str | None = None,
    email_reply_to: str | None = None,
    host: str | None = None,
    port: int | None = None,
    tls: bool | None = None,
    username: str | None = None,
    password: str | None = None,
):
    """Sends an email.

    Parameters
    ----------
    message
        The plain text message to send.
    subject
        The subject of the email.
    recipients
        The recipients of the email.
    from_address
        The email address from which the email is sent.
    html_message
        The HTML message to send.
    email_reply_to
        The email address to which to reply. Defaults to the sender.
    host
        The SMTP server host.
    port
        The SMTP server port.
    tls
        Whether to use TLS for authentication.
    username
        The SMTP server username.
    password
        The SMTP server password.

    """

    email_reply_to = email_reply_to or from_address

    msg = MIMEMultipart("alternative" if html_message else "mixed")
    msg["Subject"] = subject
    msg["From"] = from_address
    msg["To"] = ", ".join(recipients)
    msg["Reply-To"] = email_reply_to

    msg.attach(MIMEText(message, "plain"))

    if html_message:
        html = MIMEText(html_message, "html")
        msg.attach(html)

    smtp_host = host or config["notifications.smtp_server.host"]
    smtp_port = port or config["notifications.smtp_server.port"]
    smpt_tls = tls if tls is not None else config["notifications.smtp_server.tls"]
    smtp_username = username or config["notifications.smtp_server.username"]
    smtp_password = password or config["notifications.smtp_server.password"]

    with smtplib.SMTP(host=smtp_host, port=smtp_port) as smtp:
        if smpt_tls is True or (smpt_tls is None and smtp_port == 587):
            # See https://gist.github.com/jamescalam/93d915e4de12e7f09834ae73bdf37299
            smtp.ehlo()
            smtp.starttls()

            if smtp_password is not None and smtp_password is not None:
                smtp.login(smtp_username, smtp_password)
            else:
                raise ValueError("username and password must be provided for TLS.")
        smtp.sendmail(from_address, recipients, msg.as_string())


def send_critical_error_email(
    message: str,
    subject: str = "LVM Critical Alert",
    recipients: Sequence[str] | None = None,
    from_address: str | None = None,
    email_reply_to: str | None = None,
    host: str | None = None,
    port: int | None = None,
    tls: bool | None = None,
    username: str | None = None,
    password: str | None = None,
    show_critical_error_preface: bool = True,
):
    """Sends a critical error email.

    Parameters
    ----------
    message
        The message to send.
    subject
        The subject of the email.
    recipients
        The recipients of the email. A list of email addresses or :obj:`None` to
        use the default recipients.
    from_address
        The email address from which the email is sent.
    host
        The SMTP server host.
    port
        The SMTP server port.
    tls
        Whether to use TLS for authentication.
    username
        The SMTP server username.
    password
        The SMTP server password.
    show_critical_error_preface
        Whether to show a preface in the email indicating that this is a critical
        error notification.

    """

    root = pathlib.Path(__file__).parent
    template = root / config["notifications.critical.email_template"]
    loader = FileSystemLoader(template.parent)

    env = Environment(
        loader=loader,
        lstrip_blocks=True,
        trim_blocks=True,
    )
    html_template = env.get_template(template.name)

    html_message = html_template.render(
        message=message.strip(),
        show_critical_error_preface=show_critical_error_preface,
    )

    recipients = recipients or config["notifications.critical.email_recipients"]
    from_address = from_address or config["notifications.critical.email_from"]
    email_reply_to = email_reply_to or config["notifications.critical.email_reply_to"]

    assert from_address is not None, "from_address must be provided."
    assert recipients is not None, "recipients must be provided."
    assert email_reply_to is not None, "email_reply_to must be provided."

    plaintext_email = f"""A critical alert was raised in the LVM system.

The error message is shown below:

{message}

"""

    send_email(
        plaintext_email,
        subject,
        recipients,
        from_address,
        html_message=html_message,
        email_reply_to=email_reply_to,
        host=host,
        port=port,
        tls=tls,
        username=username,
        password=password,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-21
# @Filename: pubsub.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import json
import time
import uuid
from enum import auto

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    ClassVar,
    Literal,
    overload,
)

import aio_pika
from aio_pika.abc import AbstractIncomingMessage
from pydantic import BaseModel, Field
from strenum import UppercaseStrEnum
from typing_extensions import Self

from lvmopstools import config
from lvmopstools.retrier import Retrier


if TYPE_CHECKING:
    from aio_pika.abc import (
        AbstractChannel,
        AbstractConnection,
        AbstractExchange,
        AbstractQueue,
        ConsumerTag,
    )


SubCallbackType = Callable[["Message"], Awaitable[None]]
MessageType = Literal["event", "notification", "custom"]


class Event(UppercaseStrEnum):
    """Enumeration with the event types."""

    ERROR = auto()
    RECIPE_START = auto()
    RECIPE_END = auto()
    RECIPE_FAILED = auto()
    OBSERVER_NEW_TILE = auto()
    OBSERVER_STAGE_RUNNING = auto()
    OBSERVER_STAGE_DONE = auto()
    OBSERVER_STAGE_FAILED = auto()
    OBSERVER_ACQUISITION_START = auto()
    OBSERVER_ACQUISITION_DONE = auto()
    OBSERVER_STANDARD_ACQUISITION_FAILED = auto()
    DOME_OPENING = auto()
    DOME_OPEN = auto()
    DOME_CLOSING = auto()
    DOME_CLOSED = auto()
    EMERGENCY_SHUTDOWN = auto()
    UNEXPECTED_FIBSEL_REHOME = auto()
    UNCATEGORISED = auto()


class PublishedMessageModel(BaseModel):
    """A model for messages published to the exchange."""

    message_type: MessageType
    event_name: int | str | None
    payload: dict[str, Any] = {}
    timestamp: float = Field(default_factory=time.time)


class EventModel(PublishedMessageModel):
    """A model for event messages."""

    event_name: str
    message_type: Literal["event"] = "event"


class Message:
    """A model for messages to be published to the exchange."""

    def __init__(self, message: AbstractIncomingMessage):
        self.message = message

        self.body: dict[str, Any] = json.loads(message.body)
        self.payload: dict[str, Any] = self.body.get("payload", {})

        self.message_type: MessageType = self.body.get("message_type", "custom")

        self.event: Event | None = None
        self.event_name: str | None = None

        if self.message_type == "event":
            self.event_name = self.body["event_name"].upper()
            try:
                self.event = Event(self.event_name)
            except ValueError:
                self.event = Event.UNCATEGORISED


def callback_wrapper(func: SubCallbackType):
    """Wraps a callback to receive a ``Message`` instance."""

    async def wrapper(message: AbstractIncomingMessage):
        async with message.process():
            await func(Message(message))

    return wrapper


class BasePubSub:
    """A base class to connect to a RabbitMQ exchange.

    Parameters
    ----------
    connection_string
        The connection string to the RabbitMQ server.
    exchange_name
        The name of the exchange where the messages will be sent.

    """

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
    ):
        psc = config["pubsub"]

        self.connection_string = connection_string or psc["connection_string"]
        self.exchange_name = exchange_name or psc["exchange_name"]

        self.connection: AbstractConnection | None = None
        self.channel: AbstractChannel | None = None
        self.exchange: AbstractExchange | None = None

    async def connect(self) -> Self:
        """Connects to the RabbitMQ server and declares the exchange."""

        self.connection = await aio_pika.connect_robust(self.connection_string)

        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=1)

        self.exchange = await self.channel.declare_exchange(
            self.exchange_name,
            auto_delete=True,
            type=aio_pika.ExchangeType.FANOUT,
        )

        return self

    async def disconnect(self):
        """Disconnects from the RabbitMQ server."""

        if self.channel and not self.channel.is_closed:
            await self.channel.close()

        if self.connection and not self.connection.is_closed:
            try:
                await self.connection.close()
            except Exception:
                pass

    async def __aenter__(self):
        if (
            not self.connection
            or self.connection.is_closed
            or not self.channel
            or self.channel.is_closed
        ):
            await self.connect()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disconnect()


class Publisher(BasePubSub):
    """A class to publish messages to a RabbitMQ exchange. A singleton."""

    _instance: ClassVar[Publisher]

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(Publisher, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
    ):
        if not hasattr(self, "connection"):
            super().__init__(connection_string, exchange_name)

    @Retrier(max_attempts=3, delay=0.5)
    async def publish(self, message: dict, routing_key: str | None = None):
        """Publishes a message to the exchange.

        Parameters
        ----------
        message
            The message to publish. Must be a dictionary that will be encoded
            as a JSON string.
        routing_key
            The routing key to use. If not provided, uses the default routing
            key defined in the configuration.

        """

        async with self:
            assert self.exchange, "exchange not defined."

            await self.exchange.publish(
                aio_pika.Message(body=json.dumps(message).encode()),
                routing_key=routing_key or config["pubsub.routing_key"],
            )


class Subscriber(BasePubSub):
    """A class to subscribe to messages from a RabbitMQ exchange."""

    def __init__(
        self,
        connection_string: str | None = None,
        exchange_name: str | None = None,
        callback: SubCallbackType | None = None,
        queue_name: str | None = None,
    ):
        super().__init__(
            connection_string=connection_string,
            exchange_name=exchange_name,
        )

        self.queue_name: str | None = queue_name
        self.queue: AbstractQueue | None = None
        self.callback = callback

        self.consumer_tag: ConsumerTag | None = None

    async def connect(self, queue_name: str | None = None) -> Self:
        """Connects to the exchange, declares a queue, and binds the callback.

        Parameters
        ----------
        queue_name
            The name of the queue to declare. If not provided, a random name
            will be generated (recommended).

        """

        await super().connect()

        assert self.channel, "channel not defined."
        assert self.exchange, "exchange not defined."

        self.queue_name = (
            queue_name
            or self.queue_name
            or f"{self.exchange_name}-{str(uuid.uuid4()).split('-')[-1]}"
        )

        self.queue = await self.channel.declare_queue(
            self.queue_name,
            auto_delete=True,
            exclusive=True,
        )

        await self.queue.bind(
            self.exchange,
            routing_key=config["pubsub.routing_key"],
        )

        if self.callback:
            self.consumer_tag = await self.queue.consume(
                callback_wrapper(self.callback)
            )

        return self

    async def disconnect(self):
        """Disconnects from the RabbitMQ server."""

        if self.queue:
            if self.consumer_tag:
                await self.queue.cancel(self.consumer_tag)

            if self.exchange:
                await self.queue.unbind(self.exchange)

        await super().disconnect()

    @overload
    async def get(
        self,
        decode: Literal[True] = True,
    ) -> Message: ...

    @overload
    async def get(
        self,
        decode: Literal[False],
    ) -> AbstractIncomingMessage: ...

    async def get(
        self,
        decode: bool = True,
    ) -> AbstractIncomingMessage | Message:
        """Gets the next message from the queue."""

        if not self.queue:
            raise RuntimeError("queue not defined.")

        if decode:
            return Message(await self.queue.get())
        else:
            return await self.queue.get()

    @overload
    def iterator(
        self,
        decode: Literal[True] = True,
    ) -> AsyncGenerator[Message, None]: ...

    @overload
    def iterator(
        self,
        decode: Literal[False] = False,
    ) -> AsyncGenerator[AbstractIncomingMessage, None]: ...

    async def iterator(
        self,
        decode: bool = True,
    ) -> AsyncGenerator[AbstractIncomingMessage | Message, None]:
        """Iterates over a queue and yields messages."""

        async with self as instance:
            assert instance.queue, "queue not defined."

            async with instance.queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        if decode:
                            yield Message(message)
                        else:
                            yield message


async def send_event(event: Event | str, payload: dict[str, Any] = {}):
    """Convenience function to publish an event to the exchange."""

    message = EventModel(event_name=event, payload=payload).model_dump()
    await Publisher().publish(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-11
# @Filename: ion.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import warnings

from typing import cast

from typing_extensions import TypedDict

from drift import Drift
from drift.convert import data_to_float32

from lvmopstools import config
from lvmopstools.clu import send_clu_command
from lvmopstools.devices.nps import read_outlet
from lvmopstools.retrier import Retrier


__all__ = ["read_ion_pumps", "toggle_ion_pump", "convert_pressure", "ALL"]


#: Flag to toggle all ion pumps.
ALL = "all"


def convert_pressure(volts: float):
    """Converts differential voltage to pressure in Torr."""

    # The calibration is a linear fit of the form y = mx + b
    m = 2.04545
    b = -6.86373

    log10_pp0 = m * volts + b  # log10(PPa), pressure in Pascal

    torr = 10**log10_pp0 * 0.00750062

    return torr


class IonPumpDict(TypedDict):
    """Ion pump dictionary."""

    pressure: float | None
    on: bool | None
    diff_voltage: float | None


@Retrier(max_attempts=3, delay=1)
async def _read_one_ion_controller(ion_config: dict) -> dict[str, IonPumpDict]:
    """Reads the signal and on/off status from an ion controller."""

    results: dict[str, IonPumpDict] = {}

    drift = Drift(ion_config["host"], ion_config.get("port", 502), timeout=1)

    async with drift:
        for camera, camera_config in ion_config["cameras"].items():
            signal_address = camera_config["signal_address"]
            # on_off_address = camera_config["on_off_address"]

            signal = await drift.client.read_input_registers(signal_address, count=2)
            # onoff = await drift.client.read_input_registers(on_off_address, count=1)

            registers = cast(tuple[int, int], tuple(signal.registers))

            diff_volt = data_to_float32(registers)
            pressure = convert_pressure(diff_volt)

            # onoff_status = bool(onoff.registers[0])
            onoff_status = pressure > 1e-8

            # No point in reporting a bogus pressure.
            if pressure < 1e-8:
                pressure = None

            results[camera] = {
                "pressure": float(f"{pressure:.3g}") if pressure is not None else None,
                "on": onoff_status,
                "diff_voltage": round(diff_volt, 3),
            }

    return results


async def read_ion_pumps(cameras: list[str] | None = None) -> dict[str, IonPumpDict]:
    """Reads the signal and on/off status from an ion pump.

    Parameters
    ----------
    cameras
        A list of cameras to read. If `None`, reads all cameras.

    """

    ion_config: list[dict] = config["devices.ion"]

    results: dict[str, IonPumpDict] = {}
    tasks: list[asyncio.Task] = []

    for ion_controller in ion_config:
        type_ = ion_controller.get("type", "ion_controller")
        controller_cameras = ion_controller["cameras"]

        if cameras is not None:
            # Skip reading this controller if none of the cameras are in the list.
            if len(set(cameras) & set(controller_cameras)) == 0:
                continue

        for camera in controller_cameras:
            results[camera] = {"pressure": None, "on": None, "diff_voltage": None}

            if type_ == "nps":
                if (cameras is None) or (cameras is not None and camera in cameras):
                    nps_data = await read_outlet(
                        ion_controller["actor"],
                        ion_controller["outlet"],
                    )
                    results[camera] = {
                        "pressure": None,
                        "on": nps_data["state"],
                        "diff_voltage": None,
                    }

        if type_ == "ion_controller":
            tasks.append(asyncio.create_task(_read_one_ion_controller(ion_controller)))

    await asyncio.gather(*tasks, return_exceptions=True)

    for task in tasks:
        if task.exception():
            warnings.warn(f"Error reading ion pump: {task.exception()}")
            continue

        for camera, item in task.result().items():
            results[camera] = item

    if cameras is not None:
        results = {camera: results[camera] for camera in cameras if camera in results}

    if cameras is not None and set(results.keys()) != set(cameras):
        warnings.warn("Not all cameras were found in the configuration.")

    return results


@Retrier(max_attempts=3, delay=1)
async def toggle_ion_pump(camera: str, on: bool):
    """Turns the ion pump on or off.

    Parameters
    ----------
    camera
        The camera for which to toggle the ion pump. Can also be :obj:`.ALL` to
        toggle all ion pumps.
    on
        If `True`, turns the pump on. If `False`, turns the pump off. If `None`,
        toggles the pump current status.

    """

    ion_config: list[dict] = config["devices.ion"]

    if camera == ALL:
        cameras = [camera for ic in ion_config for camera in ic["cameras"]]
        for camera in cameras:
            await toggle_ion_pump(camera, on)
        return

    # ion_controller or nps
    type_: str | None = None

    # Ion pump connected to an NPS
    actor: str | None = None
    outlet: str | int | None = None

    # Ion controller box
    host: str | None = None
    port: int | None = None
    on_off_address: int | None = None

    for ic in ion_config:
        if camera in ic["cameras"]:
            type_ = ic.get("type", "ion_controller")
            if type_ == "ion_controller":
                host = ic["host"]
                port = ic.get("port", 502)
                on_off_address = ic["cameras"][camera]["on_off_address"]
            elif type_ == "nps":
                actor = ic["actor"]
                outlet = ic["outlet"]
            else:
                raise ValueError(f"Unknown type {type_!r} for ion controller {ic!r}.")

            break

    if type_ is None:
        raise ValueError(f"Camera {camera!r} not found in the configuration.")

    if type_ == "nps":
        cmd = await send_clu_command(
            f"{actor} {'on' if on else 'off'} {outlet}",
            raw=True,
        )
        if cmd.status.did_fail:
            raise ValueError(f"Error toggling ion pump for camera {camera!r} via NPS.")
        return

    # Ion controller box
    if host is None or port is None or on_off_address is None:
        raise ValueError(f"Camera {camera!r} configuration is incomplete.")

    drift = Drift(host, port)

    async with drift:
        value = 2**16 - 1 if on else 0
        await drift.client.write_register(on_off_address, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-18
# @Filename: ds9.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
import re

from typing import TYPE_CHECKING

from clu import AMQPClient, AMQPReply


if TYPE_CHECKING:
    from pyds9 import DS9


CAMERAS = [
    "sci.west",
    "sci.east",
    "skye.west",
    "skye.east",
    "skyw.west",
    "skyw.east",
    "spec.east",
]


async def ds9_agcam_monitor(
    amqp_client: AMQPClient,
    cameras: list[str] | None = None,
    replace_path_prefix: tuple[str, str] | None = None,
):
    """Shows guider images in DS9."""

    images_handled = set([])

    # Clear all frames and get an instance of DS9.
    ds9 = ds9_clear_frames()

    if cameras is None:
        cameras = CAMERAS.copy()

    if len(cameras) == 0:
        raise ValueError("No cameras specified.")

    camera_to_filename: dict[str, str | pathlib.Path | None] = {}
    for cam in cameras:
        if cam not in CAMERAS:
            raise ValueError(f"Invalid camera {cam!r}. Valid cameras: {CAMERAS!r}.")
        camera_to_filename[cam] = None

    agcam_actors = set(["lvm." + (cam.split(".")[0]) + ".agcam" for cam in cameras])

    async def handle_reply(reply: AMQPReply):
        sender = reply.sender
        if sender not in agcam_actors:
            return

        if "filename" not in reply.body:
            return

        filename: str = reply.body["filename"]["filename"]
        if filename in images_handled:
            return

        images_handled.add(filename)

        if replace_path_prefix is not None:
            filename = filename.replace(replace_path_prefix[0], replace_path_prefix[1])

        telescope = sender.split(".")[1]
        camera = reply.body["filename"]["camera"]

        is_first_all = all([vv is None for vv in camera_to_filename.values()])
        is_first_camera = camera_to_filename[f"{telescope}.{camera}"] is None

        camera_to_filename[f"{telescope}.{camera}"] = filename

        ds9_display_frames(
            camera_to_filename,
            order=cameras,
            ds9=ds9,
            show_all_frames=False,
            preserve_frames=True,
            adjust_scale=is_first_camera,
            adjust_zoom=is_first_camera,
            show_tiles=is_first_all,
        )

    amqp_client.add_reply_callback(handle_reply)

    while True:
        await asyncio.sleep(1)


def ds9_clear_frames(ds9: DS9 | None = None, ds9_target: str = "DS9:*"):
    """Clears all frames in DS9."""

    try:
        import pyds9
    except ImportError:
        raise ImportError(
            "pyds9 is not installed. You can install it manually or run"
            "pip install lvmopstools[ds9]"
        )

    if ds9 is None:
        ds9 = pyds9.DS9(target=ds9_target)

    ds9.set("frame delete all")

    return ds9


def ds9_display_frames(
    files: list[str | pathlib.Path] | dict[str, str | pathlib.Path | None],
    ds9: DS9 | None = None,
    order=CAMERAS,
    ds9_target: str = "DS9:*",
    show_all_frames=True,
    preserve_frames=True,
    clear_frames=False,
    adjust_zoom=True,
    adjust_scale=True,
    show_tiles=True,
):
    """Displays a series of images in DS9."""

    try:
        import pyds9
    except ImportError:
        raise ImportError(
            "pyds9 is not installed. You can install it manually or run"
            "pip install lvmopstools[ds9]"
        )

    if ds9 is None:
        ds9 = pyds9.DS9(target=ds9_target)

    if clear_frames:
        ds9.set("frame delete all")

    files_dict: dict[str, str | None] = {}
    if not isinstance(files, dict):
        for file_ in files:
            tel_cam = parse_agcam_filename(file_)
            if tel_cam is None:
                raise ValueError(f"Cannot parse type of file {file_!s}.")
            files_dict[".".join(tel_cam)] = str(file_)
    else:
        files_dict = {k: str(v) if v is not None else None for k, v in files.items()}

    nframe = 1
    for cam in order:
        if cam in files_dict:
            ds9.set(f"frame {nframe}")

            has_file = ds9.get("file") != ""
            zoom = ds9.get("zoom")

            if has_file:
                ds9.set("preserve pan yes")
                ds9.set("preserve regions yes")
                ds9.set("preserve scale yes")

            file_ = files_dict[cam]
            if file_ is None:
                continue

            ds9.set(f"fits {file_}")

            if adjust_scale:
                ds9.set("scale log")
                ds9.set("scale minmax")

            if adjust_zoom:
                ds9.set("zoom to fit")
            elif has_file:
                ds9.set(f"zoom {zoom}")

            nframe += 1
        else:
            if show_all_frames:
                if preserve_frames is False:
                    ds9.set(f"frame {nframe}")
                    ds9.set("frame clear")
                nframe += 1

    if show_tiles:
        ds9.set("tile")

    return ds9


def parse_agcam_filename(file_: str | pathlib.Path):
    """Returns the type of an ``agcam`` file in the form ``(telescope, camera)``."""

    file_ = pathlib.Path(file_)
    basename = file_.name

    match = re.match(".+(sci|spec|skyw|skye).+(east|west)", basename)
    if not match:
        return None

    return match.groups()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-11
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from .ion import *
from .nps import *
from .specs import *
from .switch import *
from .thermistors import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-12
# @Filename: thermistors.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import re

from typing import Literal, overload

import asyncudp

from lvmopstools import config
from lvmopstools.retrier import Retrier


__all__ = ["read_thermistors", "channel_to_valve"]


@overload
def channel_to_valve(reverse: Literal[False] = False) -> dict[int, str]: ...


@overload
def channel_to_valve(reverse: Literal[True] = True) -> dict[str, int]: ...


def channel_to_valve(reverse: bool = False) -> dict[int, str] | dict[str, int]:
    """Returns a mapping of thermistor channels to valve names.

    With ``reverse`` returns the inverse mapping, valve to device channel.

    """

    valve_to_channel = config["devices.thermistors.channels"]

    if reverse:
        return valve_to_channel

    channel_to_valve = {valve_to_channel[valve]: valve for valve in valve_to_channel}
    return channel_to_valve


@Retrier(max_attempts=3, delay=1)
async def read_thermistors():
    """Returns the thermistor values."""

    th_config = config["devices.thermistors"]

    host = th_config["host"]
    port = th_config["port"]

    socket: asyncudp.Socket | None = None

    try:
        socket = await asyncio.wait_for(
            asyncudp.create_socket(remote_addr=(host, port)),
            timeout=5,
        )

        socket.sendto(b"$016\r\n")
        data, _ = await asyncio.wait_for(socket.recvfrom(), timeout=10)
    finally:
        if socket:
            socket.close()

    match = re.match(rb"!01([0-9A-F]+)\r", data)
    if match is None:
        raise ValueError(f"Invalid response from thermistor server at {host!r}.")

    value = int(match.group(1), 16)

    thermistor_values: dict[str, bool] = {}
    for thermistor in th_config["channels"]:
        channel = th_config["channels"][thermistor]
        thermistor_values[thermistor] = bool(int((value & 1 << channel) > 0))

    return thermistor_values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-17
# @Filename: socket.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field

from typing import Any, Awaitable, Callable

from lvmopstools.retrier import Retrier


__all__ = ["AsyncSocketHandler"]

RequestFuncType = Callable[[asyncio.StreamReader, asyncio.StreamWriter], Awaitable[Any]]


@dataclass
class AsyncSocketHandler:
    """Handles a socket connection and disconnection.

    Handles secure connection and disconnection to a TCP server and executed a
    callback. By default :obj:`.Retrier` is used to retry the connection if it
    fails either during the connection phase or during callback execution.

    There are two ways to use this class. The first one is to create an instance
    and call it with a callback function which receives :obj:`~asyncio.StreamReader`
    and :obj:`~asyncio.StreamWriter` arguments ::

            async def callback(reader, writer):
                ...

            handler = AsyncSocketHandler(host, port)
            await handler(callback)

    Alternatively, you can subclass ``AsyncSocketHandler`` and override the
    :obj:`request` method ::

        class MyHandler(AsyncSocketHandler):
            async def request(self, reader, writer):
                ...

    Parameters
    ----------
    host
        The host that is running the server.
    port
        The port on which the server is listening.
    timeout
        The timeout for connection and callback execution.
    retry
        Whether to retry the connection/callback if they fails.
    retrier_params
        Parameters to pass to the :class:`.Retrier` instance.

    """

    host: str
    port: int
    timeout: float = 5
    retry: bool = True
    retrier_params: dict[str, Any] = field(default_factory=dict)

    async def _connect(self):
        """Connects to the socket."""

        reader, writer = await asyncio.wait_for(
            asyncio.open_connection(self.host, self.port),
            timeout=self.timeout,
        )

        return reader, writer

    async def _run(self, func: RequestFuncType | None = None):
        """Internal helper to connect to the socket and run the request."""

        if func is None:
            func = self.request

        reader, writer = await self._connect()

        try:
            return await func(reader, writer)
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass

    async def __call__(self, func: RequestFuncType | None = None):
        """Connects to the socket and runs the request function."""

        if self.retry:
            retrier = Retrier(**self.retrier_params)
            return await retrier(self._run)(func)
        else:
            return await self._run(func)

    async def request(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):  # pragma: no cover
        """Sends a request to the socket.

        If the handler is not called with a callback function, this method must be
        overridden in a subclass. It receives the :obj:`~asyncio.StreamReader` and
        :obj:`~asyncio.StreamWriter` client instances after a connection has been
        established.

        """

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-12
# @Filename: clu.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os

from typing import TYPE_CHECKING, Any, Literal, overload

from clu.client import AMQPClient

from lvmopstools import config


if TYPE_CHECKING:
    from clu.command import Command


__all__ = ["CluClient", "send_clu_command"]


class CluClient:
    """AMQP client asynchronous generator.

    Returns an object with an ``AMQPClient`` instance. The normal way to
    use it is to do ::

        async with CluClient() as client:
            await client.send_command(...)

    Alternatively one can do ::

        client = await anext(CluClient())
        await client.send_command(...)

    The asynchronous generator differs from the one in ``AMQPClient`` in that
    it does not close the connection on exit.

    This class is a singleton, which effectively means the AMQP client is reused
    during the life of the worker. The singleton can be cleared by calling
    `.clear`.

    The host and port for the connection can be passed on initialisation. Otherwise
    it will use the values in the environment variables ``RABBITMQ_HOST`` and
    ``RABBITMQ_PORT`` or the default values in the configuration file.

    """

    __initialised: bool = False
    __instance: CluClient | None = None

    def __new__(cls, host: str | None = None, port: int | None = None, **kwargs):
        if (
            cls.__instance is None
            or (host is not None and cls.__instance.host != host)
            or (port is not None and cls.__instance.port != port)
        ):
            cls.clear()

            cls.__instance = super(CluClient, cls).__new__(cls)
            cls.__instance.__initialised = False

        return cls.__instance

    def __init__(self, host: str | None = None, port: int | None = None, **kwargs):
        if self.__initialised is True:
            # Bail out if we are returning a singleton instance
            # which is already initialised.
            return

        host_default = os.environ.get("RABBITMQ_HOST", config["rabbitmq.host"])
        port_default = int(os.environ.get("RABBITMQ_PORT", config["rabbitmq.port"]))

        self.host: str = host or host_default
        self.port: int = port or port_default

        self.client = AMQPClient(host=self.host, port=self.port, **kwargs)
        self.__initialised = True

        self._lock = asyncio.Lock()

    def is_connected(self):
        """Is the client connected?"""

        connection = self.client.connection
        connected = connection.connection and not connection.connection.is_closed
        channel_closed = hasattr(connection, "channel") and connection.channel.is_closed

        if not connected or channel_closed:
            return False

        return True

    async def __aenter__(self):
        # Small delay to allow the event loop to update the
        # connection status if needed.
        await asyncio.sleep(0.05)

        async with self._lock:
            if not self.is_connected():
                await self.client.start()

        return self.client

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def __anext__(self):
        if not self.is_connected():
            await self.client.start()

        return self.client

    @classmethod
    def clear(cls):
        """Clears the current instance."""

        if cls.__instance and cls.__instance.is_connected():
            asyncio.create_task(cls.__instance.client.stop())

        cls.__instance = None
        cls.__initialised = False


# @overload
# async def send_clu_command(command_string: str) -> list[dict[str, Any]]: ...


@overload
async def send_clu_command(
    command_string: str,
    *,
    raw: Literal[False] = False,
    internal: bool = False,
) -> list[dict[str, Any]]: ...


@overload
async def send_clu_command(
    command_string: str,
    *,
    raw: Literal[True] = True,
    internal: bool = False,
) -> Command: ...


async def send_clu_command(
    command_string: str,
    *,
    raw=False,
    internal: bool = False,
) -> list[dict[str, Any]] | Command:
    """Sends a command to the actor system and returns a list of replies.

    Parameters
    ----------
    command_string
        The command to send to the actor. Must include the name of the actor.
    raw
        If `True`, returns the command. Otherwise returns a list of replies.

    Returns
    -------
    replies
        A list of replies, each one a dictionary of keyword to value. Empty
        replies (e.g., those only changing the status) are not returned. If
        ``raw=True``, the CLU command is returned after awaiting for it to
        complete or fail.

    Raises
    ------
    RuntimeError
        If the command fails, times out, or is otherwise not successful.

    """

    consumer, *rest = command_string.split(" ")

    async with CluClient() as client:
        cmd = await client.send_command(consumer, " ".join(rest), internal=internal)

    if cmd.status.did_succeed:
        if raw:
            return cmd

        replies: list[dict[str, Any]] = []
        for reply in cmd.replies:
            if len(reply.message) == 0:
                continue
            replies.append(reply.message)
        return replies

    raise RuntimeError(f"Command {command_string!r} failed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2025-03-20
# @Filename: switch.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import warnings

from lvmopstools import config


try:
    import netmiko
except ImportError:
    netmiko = None


__all__ = ["power_cycle_interface", "get_ag_poe_port_info"]


def _get_connection(enable: bool = True):
    """Gets a connection to the switch."""

    if not netmiko:
        raise ImportError("netmiko is required to power cycle the switch port.")

    HOST: str = os.getenv("LVM_SWITCH_HOST", config["devices.switch.host"])
    USERNAME: str = os.getenv("LVM_SWITCH_USERNAME", config["devices.switch.username"])
    PASSWORD: str | None = os.getenv("LVM_SWITCH_PASSWORD")
    SECRET: str | None = os.getenv("LVM_SWITCH_SECRET")

    if not PASSWORD:
        raise ValueError("$LVM_SWITCH_PASSWORD has not been set.")
    if not SECRET:
        raise ValueError("$LVM_SWITCH_SECRET has not been set.")

    lvm_switch = {
        "device_type": "netgear_prosafe",
        "host": HOST,
        "username": USERNAME,
        "password": PASSWORD,
        "secret": SECRET,  # Enable password
    }

    connection = netmiko.ConnectHandler(**lvm_switch)

    if enable:
        connection.enable()  # Enable method

    return connection


def power_cycle_interface(interface: str, verbose: bool = False):
    """Power cycles the PoE for a switch interface.

    The environment variales ``LVM_SWITCH_HOST``, ``LVM_SWITCH_USERNAME``,
    ``LVM_SWITCH_PASSWORD``, and ``LVM_SWITCH_SECRET`` must be set when
    calling this function.

    Parameters
    ----------
    interface
        The switch interface to power cycle, e.g., ``"2/0/6"``.
    verbose
        If ``True``, prints the commands that are being run.

    """

    connection = _get_connection()

    output = connection.send_config_set([f"interface {interface}", "poe reset"])
    if verbose:
        print(output)
        print("Closing connection ...")

    connection.disconnect()


def get_ag_poe_port_info(camera: str | None = None) -> dict[str, str]:
    """Returns the PoE port information for a given camera.

    Parameters
    ----------
    camera
        The name of the camera. If not provided, returns the information
        for all cameras.

    Returns
    -------
    dict
        A dictionary with the PoE port information for the camera.

    """

    connection = _get_connection()

    agcams_config = config["devices.agcams.power"]

    results: dict[str, str] = {}
    for cam_config_name, cam_config in agcams_config.items():
        if camera is None or camera.lower() == cam_config_name.lower():
            interface = cam_config["interface"]
            if interface is None:
                warnings.warn(f"Interface for camera {cam_config_name} is not set.")
                continue

            data = connection.send_command(f"show poe port info {interface}")
            assert isinstance(data, str)

            results[cam_config_name] = data

    if camera is not None and len(results) == 0:
        raise ValueError(f"No PoE port information found for camera {camera}.")

    connection.disconnect()

    return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-02
# @Filename: retrier.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import inspect
import time
import warnings
from dataclasses import dataclass, field
from functools import wraps

from typing import (
    Any,
    Awaitable,
    Callable,
    ParamSpec,
    TypeVar,
    overload,
)

from typing_extensions import Self


__all__ = ["Retrier"]

T = TypeVar("T", bound=Any)
P = ParamSpec("P")


@dataclass
class Retrier:
    """A class that implements a retry mechanism.

    The object returned by this class can be used to wrap a function that
    will be retried ``max_attempts`` times if it fails::

        def test_function():
            ...

        retrier = Retrier(max_attempts=5)
        retrier(test_function)()

    where the wrapped function can be a coroutine, in which case the wrapped function
    will also be a coroutine.

    Most frequently this class will be used as a decorator::

        @Retrier(max_attempts=4, delay=0.1)
        async def test_function(x, y):
            ...

        await test_function(1, 2)

    Parameters
    ----------
    max_attempts
        The maximum number of attempts before giving up.
    delay
        The delay between attempts, in seconds.
    use_exponential_backoff
        Whether to use exponential backoff for the delay between attempts. If
        :obj:`True`, the delay will be
        ``delay * exponential_backoff_base ** (attempt - 1) + random_ms`` where
        ``random_ms`` is a random number between 0 and 100 ms used to avoid
        synchronisation issues.
    exponential_backoff_base
        The base for the exponential backoff.
    max_delay
        The maximum delay between attempts when using exponential backoff.
    on_retry
        A function that will be called when a retry is attempted. The function
        should accept an exception as its only argument.
    raise_on_exception_class
        A list of exception classes that will cause an exception to be raised
        without retrying.
    timeout
        If defined, each attempt can take at most this amount of time. If the
        attempt times out, an :obj:`asyncio.TimeoutError` will be raised.
        This only works if the wrapped function is a coroutine.

    """

    max_attempts: int = 3
    delay: float = 1
    use_exponential_backoff: bool = True
    exponential_backoff_base: float = 2
    max_delay: float = 32.0
    on_retry: Callable[[Exception], None] | None = None
    raise_on_exception_class: list[type[Exception]] = field(default_factory=list)
    timeout: float | None = None

    def calculate_delay(self, attempt: int) -> float:
        """Calculates the delay for a given attempt."""

        # Random number between 0 and 100 ms to avoid synchronisation issues.
        random_ms = 0.1 * (time.time() % 1)

        if self.use_exponential_backoff:
            return min(
                self.delay * self.exponential_backoff_base ** (attempt - 1) + random_ms,
                self.max_delay,
            )
        else:
            return self.delay

    @overload
    def __call__(
        self: Self,
        func: Callable[P, T],
    ) -> Callable[P, T]: ...

    @overload
    def __call__(
        self: Self,
        func: Callable[P, Awaitable[T]],
    ) -> Callable[P, Awaitable[T]]: ...

    def __call__(
        self,
        func: Callable[P, T] | Callable[P, Awaitable[T]],
    ) -> Callable[P, T] | Callable[P, Awaitable[T]]:
        """Wraps a function to retry it if it fails."""

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs):
                attempt = 0
                while True:
                    try:
                        return await asyncio.wait_for(
                            func(*args, **kwargs),
                            timeout=self.timeout,
                        )
                    except Exception as ee:
                        attempt += 1
                        if attempt >= self.max_attempts:
                            raise ee
                        elif isinstance(ee, tuple(self.raise_on_exception_class)):
                            raise ee
                        else:
                            if self.on_retry:
                                self.on_retry(ee)
                            await asyncio.sleep(self.calculate_delay(attempt))

            return async_wrapper

        else:

            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs):
                attempt = 0
                while True:
                    try:
                        if self.timeout is not None:
                            warnings.warn(
                                "The wrapped function is not a coroutine. "
                                "The timeout parameter will be ignored.",
                                RuntimeWarning,
                            )
                        return func(*args, **kwargs)
                    except Exception as ee:
                        attempt += 1
                        if attempt >= self.max_attempts:
                            raise ee
                        elif isinstance(ee, tuple(self.raise_on_exception_class)):
                            raise ee
                        else:
                            if self.on_retry:
                                self.on_retry(ee)
                            time.sleep(self.calculate_delay(attempt))

            return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: weather.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import time
import warnings

from typing import Literal

import httpx
import polars


__all__ = ["get_weather_data", "is_weather_data_safe"]


WEATHER_URL = "http://env-api.lco.cl/metrics/weather"

WEATHER_SCHEMA = polars.Schema(
    {
        "ts": polars.String,
        "temperature": polars.Float32,
        "air_pressure": polars.Float32,
        "wind_dir_avg": polars.Float32,
        "wind_dir_max": polars.Float32,
        "wind_dir_min": polars.Float32,
        "rain_intensity": polars.Float32,
        "wind_speed_avg": polars.Float32,
        "wind_speed_max": polars.Float32,
        "wind_speed_min": polars.Float32,
        "relative_humidity": polars.Float32,
    }
)


def format_time(time: str | float) -> str:
    """Formats a time string for the LCO weather API format"""

    if isinstance(time, (float, int)):
        time = (
            datetime.datetime.fromtimestamp(
                time,
                tz=datetime.timezone.utc,
            )
            .isoformat(timespec="seconds")
            .replace("+00:00", "")
        )

    if "." in time:
        time = time.split(".")[0]

    return time


async def get_from_lco_api(
    start_time: str,
    end_time: str,
    station: str,
):  # pragma: no cover
    """Queries the LCO API for weather data."""

    start_time_dt = datetime.datetime.strptime(
        start_time,
        "%Y-%m-%dT%H:%M:%S",
    )
    end_time_dt = datetime.datetime.strptime(
        end_time,
        "%Y-%m-%dT%H:%M:%S",
    )

    data_chunks: list[polars.DataFrame] = []

    async with httpx.AsyncClient() as client:
        dt0 = start_time_dt

        # The API will only return one hour of data so we loop over it and concatenate.
        while True:
            # Use chunks of 30 minutes to be completely sure the API will return
            # the data for that interval.
            dt1 = dt0 + datetime.timedelta(hours=0.5)

            response = await client.get(
                WEATHER_URL,
                params={
                    "start_ts": dt0.strftime("%Y-%m-%dT%H:%M:%S"),
                    "end_ts": dt1.strftime("%Y-%m-%dT%H:%M:%S"),
                    "source": station.lower(),
                },
            )

            if response.status_code != 200:
                raise ValueError(f"Failed to get weather data: {response.text}")

            data = response.json()

            if "Error" in data:
                raise ValueError(f"Failed to get weather data: {data['Error']}")
            elif "results" not in data or data["results"] is None:
                raise ValueError("Failed to get weather data: no results found.")
            elif len(data["results"]) == 0:
                data_df = polars.DataFrame([], schema=WEATHER_SCHEMA)
            else:
                data_df = polars.DataFrame(data["results"], schema=WEATHER_SCHEMA)

            data_df = data_df.with_columns(
                ts=polars.col.ts.str.to_datetime(
                    time_unit="ms",
                    time_zone="UTC",
                )
            )

            data_chunks.append(data_df)

            if dt1 >= end_time_dt:
                break

            # Increase the time window by a bit less than 30 minutes to ensure that
            # we don't lose any data points.
            dt0 = dt0 + datetime.timedelta(hours=0.45)

    data = polars.concat(data_chunks).filter(
        polars.col.ts.dt.replace_time_zone(None) >= start_time_dt,
        polars.col.ts.dt.replace_time_zone(None) <= end_time_dt,
    )

    # Temperature is in Fahrenheit, convert to Celsius.
    data = data.with_columns(temperature=(polars.col.temperature - 32) * 5 / 9)

    return data


async def get_weather_data(
    start_time: str | float,
    end_time: str | float | None = None,
    station: Literal["dupont", "swope", "magellan"] = "dupont",
):
    """Returns a data frame with weather data from the du Pont station.

    .. warning::
        This function can only be run from inside the LVM network since it uses the
        LVM API which is not publicly accessible.

    Parameters
    ----------
    start_time
        The start time of the query. Can be a UNIX timestamp or an ISO datetime string.
    end_time
        The end time of the query. Can be a UNIX timestamp or an ISO datetime string.
        Defaults to the current time.
    station
        The station to query. Must be one of 'DuPont', 'C40', or 'Magellan'.

    Returns
    -------
    weather_data
        A data frame with the weather data.

    """

    if station not in ["dupont", "swope", "magellan"]:
        raise ValueError("station must be one of 'dupont', 'swope', or 'magellan'.")

    start_time = format_time(start_time)
    end_time = format_time(end_time or time.time())

    df = await get_from_lco_api(start_time, end_time, station)
    df = df.with_columns(station=polars.lit(station, polars.String))

    # Delete rows with all null values.
    df = df.filter(~polars.all_horizontal(polars.exclude("ts", "station").is_null()))

    # Sort by timestamp and keep only unique timestamps.
    df = df.sort("ts").unique("ts")

    # Calculate rolling means for average wind speed and gusts every 5m, 10m, 30m
    window_sizes = ["5m", "10m", "30m"]
    wind_cols = df.select(["wind_speed_avg", "wind_speed_max", "wind_dir_avg"])

    if wind_cols.null_count().sum_horizontal().item() > 0:
        # Sometimes we don't have wind data available, especially
        # if the anemometer has frozen.
        for ws in window_sizes:
            df = df.with_columns(
                **{
                    f"wind_speed_avg_{ws}": polars.lit(None, polars.Float32),
                    f"wind_gust_{ws}": polars.lit(None, polars.Float32),
                    f"wind_dir_avg_{ws}": polars.lit(None, polars.Float32),
                }
            )
    else:
        df = df.with_columns(
            **{
                f"wind_speed_avg_{ws}": polars.col.wind_speed_avg.rolling_mean_by(
                    by="ts",
                    window_size=ws,
                )
                for ws in window_sizes
            },
            **{
                f"wind_gust_{ws}": polars.col.wind_speed_max.rolling_max_by(
                    by="ts",
                    window_size=ws,
                )
                for ws in window_sizes
            },
            **{
                f"wind_dir_avg_{ws}": polars.col.wind_dir_avg.rolling_mean_by(
                    by="ts",
                    window_size=ws,
                )
                for ws in window_sizes
            },
        )

    # Add simple dew point.
    df = df.with_columns(
        dew_point=polars.col.temperature
        - ((100 - polars.col.relative_humidity) / 5.0).round(2)
    )

    # Change float precision to f32
    df = df.with_columns(polars.selectors.float().cast(polars.Float32))

    return df


def is_weather_data_safe(
    data: polars.DataFrame,
    measurement: str,
    threshold: float,
    now: float | datetime.datetime | None = None,
    window: int = 30,
    rolling_average_window: int = 10,
    reopen_value: float | None = None,
):
    """Determines whether an alert should be raised for a given weather measurement.

    An alert will be issued if the rolling average value of the ``measurement``
    (a column in ``data``) over the last ``window`` minutes is above the
    ``threshold``. Once the alert has been raised  the value of the ``measurement``
    must fall below the ``reopen_value`` to close the alert (defaults to the same
    ``threshold`` value) in a rolling.

    If ``now`` is not provie the current time is used as the reference point.

    ``window`` and ``rolling_average_window`` are in minutes.

    Examples
    --------
    >>> is_weather_data_safe(data, "wind_speed_avg", 35)
    True

    Returns
    -------
    result
        A boolean indicating whether the measurement is safe. `True` means the
        measurement is in a valid, safe range.

    """

    if measurement not in data.columns:
        raise ValueError(f"Measurement {measurement} not found in data.")

    reopen_value = reopen_value or threshold

    data = data.select(polars.col.ts, polars.col(measurement))
    data = data.filter(~polars.all_horizontal(polars.exclude("ts").is_null()))
    data = data.sort("ts")
    data = data.with_columns(
        timestamp=polars.col.ts.dt.timestamp("ms") / 1000,
        average=polars.col(measurement).rolling_mean_by(
            by="ts",
            window_size=f"{rolling_average_window}m",
        ),
    )

    # Get data from the last window`.
    if now is None:
        now = time.time()
    elif isinstance(now, datetime.datetime):
        if now.tzinfo is None:
            now = now.replace(tzinfo=datetime.timezone.utc)
        now = now.timestamp()

    data_window = data.filter(polars.col.timestamp > (now - window * 60))

    if data_window.height == 0:
        warnings.warn("No data available for the specified measurement.", UserWarning)
        return False

    # If any of the values in the last "window" is above the threshold, it's unsafe.
    if (data_window["average"] >= threshold).any():
        return False

    # If all the values in the last "window" are below the reopen threshold, it's safe.
    if (data_window["average"] < reopen_value).all():
        return True

    # The last case is if the values in the last "window" are between the reopen and
    # the threshold values. We want to avoid the returned value flipping from true
    # to false in a quick manner. We check the previous "window" minutes to see if
    # the alert was raised at any point. If so, we require the current window to
    # be below the reopen value. Otherwise, we consider it's safe.

    prev_window = data.filter(
        polars.col.timestamp > (now - 2 * window * 60),
        polars.col.timestamp < (now - window * 60),
    )
    if (prev_window["average"] >= threshold).any():
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-18
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import click

from sdsstools.daemonizer import cli_coro


@click.group()
def lvmopstools():
    """LVM operations tools."""

    pass


@lvmopstools.command()
@click.argument("CAMERAS", type=str, nargs=-1)
@cli_coro()
async def monitor_agcam(cameras: list[str]):
    """Monitors AGCam images in DS9."""

    from lvmopstools.ds9 import ds9_agcam_monitor
    from lvmopstools.utils import get_amqp_client

    client = await get_amqp_client()

    await ds9_agcam_monitor(client, cameras=cameras)


if __name__ == "__main__":
    lvmopstools()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-20
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio
import enum
import sys
import time
from dataclasses import dataclass

from typing import Any

import click

from clu.actor import AMQPActor
from clu.command import Command
from clu.parsers.click import CluCommand
from sdsstools import cancel_task

from lvmopstools.utils import get_exception_data


__all__ = [
    "LVMActor",
    "CheckError",
    "ActorState",
    "ErrorCodes",
    "ErrorCodesBase",
    "ErrorData",
    "create_error_codes",
]


STATE_SCHEMA = {
    "type": "object",
    "properties": {
        "code": {"type": "integer"},
        "flags": {"type": "array", "items": {"type": "string"}},
        "error": {
            "oneOf": [
                {
                    "type": "object",
                    "properties": {
                        "code": {"type": "integer"},
                        "critical": {"type": "boolean"},
                        "description": {
                            "oneOf": [{"type": "string"}, {"type": "null"}]
                        },
                        "exception": {"oneOf": [{"type": "object"}, {"type": "null"}]},
                    },
                    "required": ["code", "critical", "description"],
                    "additionalProperties": False,
                },
                {"type": "null"},
            ]
        },
        "additionalProperties": False,
    },
}


class ActorState(enum.Flag):
    """Defines the possible states of the actor."""

    RUNNING = 1 << 0
    READY = 1 << 1
    TROUBLESHOOTING = 1 << 2
    TROUBLESHOOT_FAILED = 1 << 3
    RESTARTING = 1 << 4
    CHECKING = 1 << 5

    NOT_READY = TROUBLESHOOTING | TROUBLESHOOT_FAILED | RESTARTING
    SKIP_CHECK = CHECKING | TROUBLESHOOTING | RESTARTING

    def get_state_names(self):
        """Returns the state names that are set."""

        return [state.name for state in self.__class__ if self & state]


@dataclass
class ErrorData:
    code: int
    critical: bool = False
    description: str = ""


class ErrorCodesBase(enum.Enum):
    """Enumeration of error codes"""

    @classmethod
    def get_error_code(cls, error_code: int):
        """Returns the :obj:`.ErrorCodes` that matches the ``error_code`` value."""

        for error in cls:
            if error.value.code == error_code:
                return error

        raise ValueError(f"Error code {error_code} not found.")


def create_error_codes(
    error_codes: dict[str, tuple | list | ErrorData],
    name: str = "ErrorCodes",
    include_unknown: bool = True,
) -> Any:
    """Creates an enumeration of error codes."""

    error_codes_enum: dict[str, ErrorData] = {}
    for error_name, error_data in error_codes.items():
        if not isinstance(error_data, ErrorData):
            error_data = ErrorData(*error_data)
        error_codes_enum[error_name.upper()] = error_data

    if include_unknown and "UNKNOWN" not in error_codes_enum:
        error_codes_enum["UNKNOWN"] = ErrorData(9999, True, "Unknown error")

    return ErrorCodesBase(name, error_codes_enum)


ErrorCodes = create_error_codes({"UNKNOWN": ErrorData(9999, True, "Unknown error")})


@click.command(cls=CluCommand, name="actor-state")
async def actor_state(command: Command[LVMActor], *args, **kwargs):
    """Returns the actor state."""

    state = command.actor.state
    code = state.value
    flags = state.get_state_names()

    if (model := command.actor.model) is not None:
        state_kw = model["state"]
        if state_kw is not None and state_kw.value:
            state = state_kw.value.copy()
            state.update({"code": code, "flags": flags})
            return command.finish(state=state)

    return command.finish(state={"code": code, "flags": flags, "error": None})


@click.command(cls=CluCommand, name="actor-restart")
@click.option(
    "--mode",
    "-m",
    type=click.Choice(["exit", "reload"]),
    default="exit",
    help="How to restart the actor.",
)
async def actor_restart(command: Command[LVMActor], mode: str = "exit"):
    """Restarts the actor."""

    await command.actor.restart(mode=mode)
    return command.finish()


class CheckError(Exception):
    """An exception raised when the :obj:`.LVMActor` check fails."""

    def __init__(
        self,
        message: str = "",
        error_code: ErrorCodesBase | int = ErrorCodes.UNKNOWN,
    ):
        if isinstance(error_code, int):
            self.error_code = ErrorCodes.get_error_code(error_code)
        else:
            self.error_code = error_code

        self.message = message

        super().__init__(message)


class LVMActor(AMQPActor):
    """Base class for LVM actors."""

    def __init__(
        self,
        *args,
        check_interval: float = 30.0,
        restart_after: float | None = 300.0,
        restart_mode="reload",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Actor state.
        self.state = ActorState(0)
        self.check_interval = check_interval
        self._check_task: asyncio.Task | None = None
        self._last_not_ready: float = -1

        self.restart_after = restart_after
        self.restart_mode = restart_mode

        # Additional commands.
        self.parser.add_command(actor_state)
        self.parser.add_command(actor_restart)

        # Add keywords in schema for the actor state.
        assert self.model and self.model.schema, "Model schema not defined"
        self.model.schema["properties"]["state"] = STATE_SCHEMA

    async def start(self):  # pragma: no cover
        """Starts the actor."""

        await super().start()
        self.update_state(ActorState.RUNNING)

        self._check_task = asyncio.create_task(self._check_loop())

        return self

    async def stop(self):
        """Stops the actor."""

        await cancel_task(self._check_task)
        await self.timed_commands.stop()

        return await super().stop()

    async def _check_loop(self):
        """Runs the check loop."""

        while True:
            if self.state & ActorState.SKIP_CHECK:
                await asyncio.sleep(self.check_interval)
                continue

            if not self.is_ready() and self._last_not_ready > 0:
                not_ready_time = time.time() - self._last_not_ready
                if self.restart_after and not_ready_time > self.restart_after:
                    self.write(
                        "w",
                        text=f"It has been {not_ready_time:.0f} seconds since the "
                        "actor was last ready. Restarting.",
                    )
                    await self.restart()

            try:
                self.state |= ActorState.CHECKING
                await self._check_internal()
            except CheckError as err:
                await self.troubleshoot(error_code=err.error_code, exception=err)
            except Exception as err:
                await self.troubleshoot(exception=err, traceback_frame=1)
            else:
                self.update_state(ActorState.READY)
            finally:
                await asyncio.sleep(self.check_interval)

    def is_ready(self):
        """Returns :obj:`True` if the actor is ready."""

        return bool(self.state & ActorState.READY)

    def update_state(
        self,
        state: ActorState,
        error_data: dict[str, Any] | None = None,
        command: Command | None = None,
        internal: bool = True,
    ):
        """Updates the state and broadcasts the change."""

        old_state = self.state

        self.state = ActorState(0)
        self.state |= state

        if self.state & ActorState.NOT_READY:
            self.state &= ~ActorState.READY
            if self._last_not_ready < 0:
                self._last_not_ready = time.time()
        else:
            self._last_not_ready = -1

        if self.is_connected():
            self.state |= ActorState.RUNNING

        if old_state.value != self.state.value:
            self.write(
                "d",
                state={
                    "code": self.state.value,
                    "flags": self.state.get_state_names(),
                    "error": error_data,
                },
                internal=internal,
                command=command,
            )

        return self.state

    async def troubleshoot(
        self,
        error_code: ErrorCodesBase = ErrorCodes.UNKNOWN,
        exception: Exception | None = None,
        traceback_frame: int = 0,
    ):
        """Handles troubleshooting."""

        error_data = {
            "code": error_code.value.code,
            "critical": error_code.value.critical,
            "description": error_code.value.description,
            "exception": get_exception_data(exception, traceback_frame=traceback_frame),
        }

        self.update_state(ActorState.TROUBLESHOOTING, error_data=error_data)

        if await self._troubleshoot_internal(error_code, exception=exception):
            self.update_state(ActorState.READY)
        else:
            self.update_state(ActorState.TROUBLESHOOT_FAILED)

    async def restart(self, mode: str | None = None):
        """Restarts the actor by killing the process.

        Parameters
        ----------
        mode
            How to restart the actor. Possible values are ``"exit"``, which will
            finish the process and let the supervisor restart it (for example
            a Kubernetes scheduler), and ``"reload"`` which will stop and restart
            the actor without killing the process. If ``None``, defaults to
            ``restart_mode``.

        """

        mode = mode or self.restart_mode

        self.write("w", text=f"Restarting {self.name} with mode={mode!r}.")
        self.update_state(ActorState.RESTARTING)

        await asyncio.sleep(1)

        if mode == "exit":
            sys.exit(1)
        elif mode == "reload":
            try:
                await self.stop()
            finally:
                await self.start()
        else:
            raise ValueError("Invalid restart mode.")

    @abc.abstractmethod
    async def _check_internal(self):
        """Checks the status of the actor.

        This method is intended for the user to override when subclassing from
        :obj:`.LVMActor`. It is called by :obj:`.check_loop` every ``check_interval``
        seconds. The method must perform any necessary checks and, if a problem
        is found, call :obj:`.troubleshoot` with the appropriate error code.
        Alternatively if the check fails it can raise a :obj:`.CheckError` with
        the appropriate error code.

        """

        pass

    @abc.abstractmethod
    async def _troubleshoot_internal(
        self,
        error_code: ErrorCodesBase,
        exception: Exception | None = None,
    ):
        """Handles internal troubleshooting.

        This methos is intended for the user to override when subclassing from
        :obj:`.LVMActor`. It is called by :obj:`.troubleshoot` after updating
        the actor state to :obj:`.State.TROUBLESHOOTING`.

        The method must perform any necessary troubleshooting and then return
        a boolean indicating whether the actor is ready to continue or not.
        :obj:`.troubleshoot` will then update the actor state accordingly.

        """

        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-01-18
# @Filename: utils.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import subprocess
import time
import warnings
from functools import wraps

from typing import Any, Coroutine, TypeVar

import nmap3

from clu import AMQPClient
from sdsstools.utils import run_in_executor


try:
    import netmiko
except ImportError:
    netmiko = None


__all__ = [
    "get_amqp_client",
    "get_exception_data",
    "stop_event_loop",
    "with_timeout",
    "timeout",
    "is_notebook",
    "Trigger",
    "is_host_up",
    "is_root",
]


async def get_amqp_client(**kwargs) -> AMQPClient:  # pragma: no cover
    """Returns a CLU AMQP client."""

    amqp_client = AMQPClient(**kwargs)
    await amqp_client.start()

    return amqp_client


def get_exception_data(exception: Exception | None, traceback_frame: int = 0):
    """Returns a dictionary with information about an exception."""

    if exception is None:
        return None

    if not isinstance(exception, Exception):
        return None

    exception_data: dict[str, Any] = {}
    if exception is not None:
        filename: str | None = None
        lineno: int | None = None
        if exception.__traceback__ is not None:
            tb = exception.__traceback__
            for _ in range(traceback_frame):
                t_next = tb.tb_next
                if t_next is None:
                    break
                tb = t_next

            filename = tb.tb_frame.f_code.co_filename if tb else None
            lineno = tb.tb_lineno if tb else None

        exception_data = {
            "module": exception.__class__.__module__,
            "type": exception.__class__.__name__,
            "message": str(exception),
            "filename": filename,
            "lineno": lineno,
        }

    return exception_data


async def stop_event_loop(timeout: float | None = 5):  # pragma: no cover
    """Cancels all running tasks and stops the event loop."""

    for task in asyncio.all_tasks():
        task.cancel()

    try:
        await asyncio.wait_for(asyncio.gather(*asyncio.all_tasks()), timeout=timeout)
    except asyncio.TimeoutError:
        pass
    finally:
        asyncio.get_running_loop().stop()


def is_notebook() -> bool:
    """Returns :obj:`True` if the code is run inside a Jupyter Notebook.

    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    """

    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


T = TypeVar("T", bound=Any)


async def with_timeout(
    coro: asyncio.Future[T] | Coroutine[Any, Any, T],
    timeout: float | None,
    raise_on_timeout: bool = True,
) -> T | None:
    """Runs a coroutine with a timeout.

    Parameters
    ----------
    coro
        The coroutine to run.
    timeout
        The timeout in seconds.
    raise_on_timeout
        If :obj:`True`, raises a :class:`asyncio.TimeoutError` if the coroutine times
        out, otherwise returns :obj:`None`.

    Returns
    -------
    result
        The result of the coroutine.

    Raises
    ------
    asyncio.TimeoutError
        If the coroutine times out.

    """

    try:
        return await asyncio.wait_for(coro, timeout)
    except asyncio.TimeoutError:
        if raise_on_timeout:
            raise asyncio.TimeoutError(f"Timed out after {timeout} seconds.")


def timeout(timeout: float, raise_on_timeout: bool = True):
    """Decorator that wraps an async function with :obj:`asyncio.wait_for`.

    Parameters
    ----------
    timeout
        The timeout in seconds.
    raise_on_timeout
        If :obj:`True`, raises a :class:`asyncio.TimeoutError` if the coroutine times
        out, otherwise returns :obj:`None`.

    Example
    -------
    @timeout(5)
    async def work(): ...

    @timeout(5, raise_on_timeout=False)
    async def maybe(): ...

    """

    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("timeout decorator can only be applied to async functions.")

        @wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await asyncio.wait_for(func(*args, **kwargs), timeout)
            except asyncio.TimeoutError:
                if raise_on_timeout:
                    raise asyncio.TimeoutError(
                        f"{func.__name__} timed out after {timeout} seconds."
                    )
                return None

        return wrapper

    return decorator


class Trigger:
    """A trigger that can be set and reset and accepts setting thresholds.

    This class is essentially just a flag that can take true/false values, but
    triggering the true value can be delayed by time or number or triggers.

    Parameters
    ----------
    n
        The number of times the instance needs to be set before it is triggered.
    delay
        The delay in seconds before the instance is triggered. This is counted from
        the first time the instance is set, and is reset if the instance is reset.
        If ``n_triggers`` is greater than 1, both conditions must be met for the
        instance to be triggered.

    """

    def __init__(self, n: int = 1, delay: float = 0):
        self.n = n
        self.delay = delay

        self._triggered = False
        self._first_trigger: float | None = None
        self._n_sets: int = 0

    def _check(self):
        """Check the trigger conditions and update the internal state."""

        now = time.time()
        if (
            self._n_sets >= self.n
            and self._first_trigger is not None
            and now - self._first_trigger >= self.delay
        ):
            self._triggered = True

    def set(self):
        """Sets the trigger."""

        if self._triggered:
            return

        self._n_sets += 1
        self._first_trigger = self._first_trigger or time.time()
        self._check()

    def reset(self):
        """Resets the trigger."""

        self._first_trigger = None
        self._n_sets = 0
        self._triggered = False

    def is_set(self):
        """Returns :obj:`True` if the trigger is set."""

        self._check()

        return self._triggered


def is_root():
    """Returns whether the user is root. This may not work in all systems."""

    return os.geteuid() == 0


async def ping_host(host: str):
    """Pings a host."""

    cmd = await asyncio.create_subprocess_exec(
        *["ping", "-c", "1", "-W", "5", host],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    return_code = await cmd.wait()
    return return_code == 0


async def is_host_up(host: str, fallback_to_ping: bool = True) -> bool:
    """Returns whether a host is up.

    Parameters
    ----------
    host
        The host to check.
    fallback_to_ping
        If ``True``, a system ping will be used if the user is not root or nmap
        is not available. This is less reliable than using nmap, but nmap requires
        running as root for reliable results.

    Returns
    -------
    is_up
        ``True`` if the host is up, ``False`` otherwise.

    """

    if not is_root():
        if fallback_to_ping:
            warnings.warn('Running as non-root; using "ping" instead of "nmap".')
            return await ping_host(host)
        raise PermissionError("root privileges are required to run nmap.")

    if not nmap3.get_nmap_path():
        if fallback_to_ping:
            warnings.warn('nmap not available. Using "ping instead.')
            return await ping_host(host)
        raise RuntimeError("nmap is not available.")

    nmap = nmap3.NmapHostDiscovery()
    result = await run_in_executor(
        nmap.nmap_no_portscan,
        host,
        args="--host-timeout=1 --max-retries=2",
    )

    if (
        host not in result
        or not isinstance(result[host], dict)
        or "state" not in result[host]
        or "state" not in result[host]["state"]
    ):
        return False

    return result[host]["state"]["state"] == "up"
